#include <math.h>

#include "imcaXiaFiltersLib.h"

#define MAX_THICKNESS_DELTA 0.001

static void addCombination(struct ImcaXiaFiltersData *, int[], int);
static void addEmptyFilterSet(struct ImcaXiaFiltersData *, int);
static int nextCombination(int[], int, int);

int imcaXiaAreAllStatesSame(short states[], size_t statesSize) {
  int i;

  for (i = 1; i < statesSize; i++) if (states[i] != states[0]) return 0;
  return 1;
}

int imcaXiaFiltersFindNearestAttenuation(struct ImcaXiaFiltersData *data,
    double attenuation) {
  int i;
  double iDelta;
  double iMinusOneDelta;

  if (data->numFilterSets == 0) return -1;
  for (i = 0; i < data->numFilterSets; i++)
    if (attenuation <= data->filterSets[i].attenuation) break;
  if (i == 0) return i;
  if (i == data->numFilterSets) return i - 1;
  iDelta = fabs(attenuation - data->filterSets[i].attenuation);
  iMinusOneDelta = fabs(attenuation - data->filterSets[i - 1].attenuation);
  return (iDelta < iMinusOneDelta) ? i : i - 1;
}

int imcaXiaFiltersFindNearestThickness(struct ImcaXiaFiltersData *data,
    double thickness) {
  int i;
  double iDelta;
  double iMinusOneDelta;

  if (data->numFilterSets == 0) return -1;
  for (i = 0; i < data->numFilterSets; i++)
    if (thickness <= data->filterSets[i].thickness) break;
  if (i == 0) return i;
  if (i == data->numFilterSets) return i - 1;
  iDelta = fabs(thickness - data->filterSets[i].thickness);
  iMinusOneDelta = fabs(thickness - data->filterSets[i - 1].thickness);
  return (iDelta < iMinusOneDelta) ? i : i - 1;
}

void imcaXiaFiltersInitialize(struct ImcaXiaFiltersData *data,
    double filterThicknesses[], size_t filterThicknessesSize,
    double attenuationCoefficient) {
  int i, j;
  int n, r;
  int indices[IMCA_XIA_FILTERS_MAX_NUM_FILTERS];

  for (i = 0; i < IMCA_XIA_FILTERS_MAX_NUM_THICKNESSES; i++) {
    data->filterSets[i].thickness = 0.0;
    for (j = 0; j < IMCA_XIA_FILTERS_MAX_NUM_FILTERS; j++)
      data->filterSets[i].states[j] = 0;
    data->filterSets[i].attenuation = 0.0;
  }

  for (i = 0; i < filterThicknessesSize; i++)
    data->filterThicknesses[i] = filterThicknesses[i];
  for (i = filterThicknessesSize; i < IMCA_XIA_FILTERS_MAX_NUM_FILTERS; i++)
    data->filterThicknesses[i] = 0.0;
  data->numFilterSets = 1;

  n = IMCA_XIA_FILTERS_MAX_NUM_FILTERS;
  for (r = 1; r <= n; r++) {
    for (i = 0; i < r; i++) indices[i] = i;
    addCombination(data, indices, r);
    while (nextCombination(indices, n, r))
      addCombination(data, indices, r);
  }

  imcaXiaFiltersUpdateAttenuation(data, attenuationCoefficient);
}

void imcaXiaFiltersUpdateAttenuation(struct ImcaXiaFiltersData *data,
    double attenuationCoefficient) {
  int i;

  for (i = 0; i < data->numFilterSets; i++)
    data->filterSets[i].attenuation = (1.0 - (1.0 /
      exp(attenuationCoefficient * data->filterSets[i].thickness))) * 100.0;
}

static void addCombination(struct ImcaXiaFiltersData *data, int indices[],
    int r) {
  double thickness;
  int i;
  int added;
  int index;
  int oldNumInsertedStates;
  int newNumInsertedStates;

  thickness = 0.0;
  for (i = 0; i < r; i++) thickness += data->filterThicknesses[indices[i]];

  added = 0;
  for (i = 0; i < data->numFilterSets; i++) {
    if (fabs(thickness - data->filterSets[i].thickness) <=
          MAX_THICKNESS_DELTA)
      break;
    if (thickness < data->filterSets[i].thickness) {
      addEmptyFilterSet(data, i);
      added = 1;
      break;
    }
  }
  index = i;
  if (index == data->numFilterSets) {
    added = 1;
    data->numFilterSets++;
  }

  if (!added) {
    oldNumInsertedStates = 0;
    for (i = 0; i < IMCA_XIA_FILTERS_MAX_NUM_FILTERS; i++)
      if (data->filterSets[index].states[i] == 1) oldNumInsertedStates++;
    newNumInsertedStates = r;
    if (oldNumInsertedStates <= newNumInsertedStates) return;
  }

  if (added) data->filterSets[index].thickness = thickness;

  for (i = 0; i < IMCA_XIA_FILTERS_MAX_NUM_FILTERS; i++)
    data->filterSets[index].states[i] = 0;
  for (i = 0; i < r; i++)
    data->filterSets[index].states[indices[i]] = 1;
}

static void addEmptyFilterSet(struct ImcaXiaFiltersData *data, int index) {
  int i;

  i = data->numFilterSets - 1;
  if (i + 1 == IMCA_XIA_FILTERS_MAX_NUM_THICKNESSES) i--;
  for (; i >= index; i--) data->filterSets[i + 1] = data->filterSets[i];
  data->filterSets[index].thickness = 0.0;
  for (i = 0; i < IMCA_XIA_FILTERS_MAX_NUM_FILTERS; i++)
    data->filterSets[index].states[i] = 0;
  data->filterSets[index].attenuation = 0.0;
  data->numFilterSets++;
}

/*
 * Steps indices to next r-combination from the set {0, 1, 2,..., n - 1}.
 *
 * Returns 1 if stepped; 0 if there were no more combinations possible.
 *
 * Based on pseudocode from Discrete Mathematics and Its Applications, 3rd
 * Edition, by Kenneth H. Rosen, New York: McGraw-Hill, 1995, page 297.
 */
static int nextCombination(int indices[], int n, int r) {
  int i, j;

  i = r - 1;
  while (i >= 0 && indices[i] == n - r + i) i--;
  if (i < 0) return 0;
  indices[i] = indices[i] + 1;
  for (j = i + 1; j < r; j++) indices[j] = indices[i] + j - i;
  return 1;
}
