#include <math.h>

#include <iocsh.h>
#include <epicsExport.h>
#include <registryFunction.h>

#include <imca-epics-unit.h>

#include "imcaXiaFiltersLib.h"

static EpicsUnitContext euc;

static void testLibAreAllStatesSame01() {
  /* Basic. */
  short states[] = {0, 0, 0, 0, 0, 0, 0, 0};

  failIfIntNotEqual(&euc, 1, imcaXiaAreAllStatesSame(states,
    sizeof(states) / sizeof(states[0])));
}
static void testLibAreAllStatesSame02() {
  /* Basic. */
  short states[] = {1, 1, 1, 1, 1, 1, 1, 1};

  failIfIntNotEqual(&euc, 1, imcaXiaAreAllStatesSame(states,
    sizeof(states) / sizeof(states[0])));
}
static void testLibAreAllStatesSame03() {
  /* Basic. */
  short states[] = {0, 1, 1, 1, 1, 1, 1, 1};

  failIfIntNotEqual(&euc, 0, imcaXiaAreAllStatesSame(states,
    sizeof(states) / sizeof(states[0])));
}
static void testLibAreAllStatesSame04() {
  /* Basic. */
  short states[] = {1, 1, 0, 1, 1, 1, 1, 1};

  failIfIntNotEqual(&euc, 0, imcaXiaAreAllStatesSame(states,
    sizeof(states) / sizeof(states[0])));
}
static void testLibAreAllStatesSame05() {
  /* Basic. */
  short states[] = {1, 1, 1, 1, 1, 1, 1, 0};

  failIfIntNotEqual(&euc, 0, imcaXiaAreAllStatesSame(states,
    sizeof(states) / sizeof(states[0])));
}
static void testLibAreAllStatesSame06() {
  /* Basic. */
  short states[] = {1, 1, 1, 1, 0, 0, 0, 0};

  failIfIntNotEqual(&euc, 1, imcaXiaAreAllStatesSame(states, 4));
}

static void testLibInitialize01() {
  /* Basic. */
  double filterThicknesses[] = {
    10.0, 20.0, 40.0, 80.0, 150.0, 300.0, 600.0, 1200.0
  };
  struct ImcaXiaFiltersData data;
  double attenuationCoefficient;
  int thickness;
  int i;

  attenuationCoefficient = 0.004;
  imcaXiaFiltersInitialize(&data, filterThicknesses,
    sizeof(filterThicknesses) / sizeof(filterThicknesses[0]),
    attenuationCoefficient);

  failIfIntNotEqual(&euc, 241, data.numFilterSets);
  thickness = 0;
  for (i = 0; i < data.numFilterSets; i++) {
    failIfDoubleNotEqual(&euc, (double)thickness,
      data.filterSets[i].thickness, 1e-4);
    failIfDoubleNotEqual(&euc, (1.0 -
        (1.0 / exp(attenuationCoefficient * (double)thickness))) * 100.0,
      data.filterSets[i].attenuation, 1e-4);
    thickness += 10;
  }
}
static void testLibInitialize02() {
  /* Prefer inserting fewer filters when possible (e.g. 30 over 10 + 20). */
  double filterThicknesses[] = {
    10.0, 20.0, 30.0
  };
  struct ImcaXiaFiltersData data;
  int thickness;
  int i;

  imcaXiaFiltersInitialize(&data, filterThicknesses,
    sizeof(filterThicknesses) / sizeof(filterThicknesses[0]), 0.004);

  failIfIntNotEqual(&euc, 7, data.numFilterSets);
  thickness = 0;
  for (i = 0; i < data.numFilterSets; i++) {
    failIfDoubleNotEqual(&euc, (double)thickness,
      data.filterSets[i].thickness, 1e-4);
    thickness += 10;
  }
  for (i = 0; i < IMCA_XIA_FILTERS_MAX_NUM_FILTERS; i++)
    failIfIntNotEqual(&euc, (i == 2) ? 1 : 0, data.filterSets[3].states[i]);
}

static void testLibUpdateAttenuation01() {
  /* Basic. */
  double filterThicknesses[] = {
    10.0, 20.0, 40.0, 80.0, 150.0, 300.0, 600.0, 1200.0
  };
  struct ImcaXiaFiltersData data;
  double attenuationCoefficient;
  int thickness;
  int i;

  attenuationCoefficient = 0.004;
  imcaXiaFiltersInitialize(&data, filterThicknesses,
    sizeof(filterThicknesses) / sizeof(filterThicknesses[0]),
    attenuationCoefficient);

  attenuationCoefficient = 0.005;
  imcaXiaFiltersUpdateAttenuation(&data, attenuationCoefficient);
  thickness = 0;
  for (i = 0; i < data.numFilterSets; i++) {
    failIfDoubleNotEqual(&euc, (1.0 -
        (1.0 / exp(attenuationCoefficient * (double)thickness))) * 100.0,
      data.filterSets[i].attenuation, 1e-4);
    thickness += 10;
  }
}

static void runAllTests() {
  runTest(&euc, testLibAreAllStatesSame01);
  runTest(&euc, testLibAreAllStatesSame02);
  runTest(&euc, testLibAreAllStatesSame03);
  runTest(&euc, testLibAreAllStatesSame04);
  runTest(&euc, testLibAreAllStatesSame05);
  runTest(&euc, testLibAreAllStatesSame06);

  runTest(&euc, testLibInitialize01);
  runTest(&euc, testLibInitialize02);

  runTest(&euc, testLibUpdateAttenuation01);
}

void imcaXiaFiltersLibTestRunner() {
  runTests(&euc, runAllTests);
}

static const iocshArg *const imcaXiaFiltersLibTestRunnerArgs[] = {};
static const iocshFuncDef imcaXiaFiltersLibTestRunnerDef = {
  "imcaXiaFiltersLibTestRunner", 0, imcaXiaFiltersLibTestRunnerArgs
};
static void imcaXiaFiltersLibTestRunnerCall(const iocshArgBuf *args) {
  imcaXiaFiltersLibTestRunner();
}
static void imcaXiaFiltersLibTestRegistrar(void) {
  static int isFirstTime = 1;
  if (!isFirstTime) return;
  isFirstTime = 0;
  iocshRegister(&imcaXiaFiltersLibTestRunnerDef,
    imcaXiaFiltersLibTestRunnerCall);
}
epicsExportRegistrar(imcaXiaFiltersLibTestRegistrar);
