/*
 * I insert and retract the filters of two XIA PFCU-4 units combined to form a
 * logical 8-filter device to maintain the requested total thickness or
 * attenuation percentage at the present energy.
 *
 * Macro parameters:
 *   P   PV prefix (e.g. x7idb:)
 *   D   Name of device; same as D in imcaXiaPfcu4x2.db (e.g. FltrRak1)
 *   EM  Name of PV providing energy-moving flag (0=Idle, 1=Moving) (e.g.
 *       x7ida:EnergyMoving)
 */

program imcaXiaPfcu4x2
option +r;


%%#include <stdlib.h>
%%#include <epicsAssert.h>

%%#include "imcaXiaFiltersLib.h"


#define PROGRAM_NAME "imcaXiaPfcu4x2"

#define NUM_FILTERS 8


short EnergyMoving;
  assign EnergyMoving to "{EM}";
  monitor EnergyMoving;
  evflag EnergyMovingChanged;
  sync EnergyMoving EnergyMovingChanged;

short RequestedAttenuationUpdate;
  assign RequestedAttenuationUpdate to "{P}{D}:RqsAtten.PROC";

double RequestedAttenuation;
  assign RequestedAttenuation to "{P}{D}:RqsAtten";
  monitor RequestedAttenuation;
  syncq RequestedAttenuation 100;

short RequestedAttenuationBusy;
  assign RequestedAttenuationBusy to "{P}{D}:RqsAttenBusy";
  monitor RequestedAttenuationBusy;

short MaintainRequestedAttenuation;
  assign MaintainRequestedAttenuation to "{P}{D}:MntnRqsAtten";
  monitor MaintainRequestedAttenuation;
  evflag MaintainRequestedAttenuationChanged;
  sync MaintainRequestedAttenuation MaintainRequestedAttenuationChanged;

short AttenuationCoefficientUpdate;
  assign AttenuationCoefficientUpdate to "{P}{D}:CalcEnergy.PROC";

double AttenuationCoefficient;
  assign AttenuationCoefficient to "{P}{D}:AttenCoeff";

double RequestedThickness;
  assign RequestedThickness to "{P}{D}:RqsThickSum";
  monitor RequestedThickness;
  syncq RequestedThickness 100;

short RequestedThicknessBusy;
  assign RequestedThicknessBusy to "{P}{D}:RqsThickSumBusy";
  monitor RequestedThicknessBusy;

double FilterThicknesses[NUM_FILTERS];
  assign FilterThicknesses[0] to "{P}{D}:FltrThick1";
  assign FilterThicknesses[1] to "{P}{D}:FltrThick2";
  assign FilterThicknesses[2] to "{P}{D}:FltrThick3";
  assign FilterThicknesses[3] to "{P}{D}:FltrThick4";
  assign FilterThicknesses[4] to "{P}{D}:FltrThick5";
  assign FilterThicknesses[5] to "{P}{D}:FltrThick6";
  assign FilterThicknesses[6] to "{P}{D}:FltrThick7";
  assign FilterThicknesses[7] to "{P}{D}:FltrThick8";

short RequestedFilterStates[NUM_FILTERS];
  assign RequestedFilterStates[0] to "{P}{D}:Filter1Bo";
  assign RequestedFilterStates[1] to "{P}{D}:Filter2Bo";
  assign RequestedFilterStates[2] to "{P}{D}:Filter3Bo";
  assign RequestedFilterStates[3] to "{P}{D}:Filter4Bo";
  assign RequestedFilterStates[4] to "{P}{D}:Filter5Bo";
  assign RequestedFilterStates[5] to "{P}{D}:Filter6Bo";
  assign RequestedFilterStates[6] to "{P}{D}:Filter7Bo";
  assign RequestedFilterStates[7] to "{P}{D}:Filter8Bo";

short RequestedAllFiltersState;
  assign RequestedAllFiltersState to "{P}{D}:FilterAllBo";


int Index;
int AllStatesSame;
int Status;
int IgnoreNextRequestedAttenuationChange;
int IgnoreNextRequestedThicknessChange;
char *FiltersData; /* Really "struct ImcaXiaFiltersData *" */


entry {
  %%pVar->FiltersData = (char *)malloc(sizeof(struct ImcaXiaFiltersData));
  %%assert(pVar->FiltersData != NULL);
}


ss Main {
  state initialize {
    when (pvConnectCount() == pvChannelCount()) {
      IgnoreNextRequestedAttenuationChange = 0;
      IgnoreNextRequestedThicknessChange = 0;

      for (Index = 0; Index < NUM_FILTERS; Index++) {
        Status = pvGet(FilterThicknesses[Index]);
        %%assert(pVar->Status == pvStatOK);
      }
      %{
        imcaXiaFiltersInitialize(
            (struct ImcaXiaFiltersData *)pVar->FiltersData,
            pVar->FilterThicknesses, NUM_FILTERS,
            pVar->AttenuationCoefficient);
      }%

      efClear(EnergyMovingChanged);
    } state waitForSomethingToDo
  }

  state waitForSomethingToDo {
    when (pvGetQ(RequestedAttenuation) || RequestedAttenuationBusy) {
    } state handleRequestedAttenuationChanged
    when (pvGetQ(RequestedThickness) || RequestedThicknessBusy) {
    } state handleRequestedThicknessChanged
    when (efTestAndClear(EnergyMovingChanged) && !EnergyMoving) {
    } state handleEnergyMoved
    when (efTestAndClear(MaintainRequestedAttenuationChanged) &&
        MaintainRequestedAttenuation) {
    } state handleMaintainRequestedAttenuationEnabled
  }

  state handleRequestedAttenuationChanged {
    when (IgnoreNextRequestedAttenuationChange) {
      IgnoreNextRequestedAttenuationChange = 0;
      RequestedAttenuationBusy = 0;
      Status = pvPut(RequestedAttenuationBusy);
      %%assert(pVar->Status == pvStatOK);
    } state waitForSomethingToDo
    when (!IgnoreNextRequestedAttenuationChange) {
      %%struct ImcaXiaFiltersData *data;
      %%int i, idx;

      AttenuationCoefficientUpdate = 1;
      Status = pvPut(AttenuationCoefficientUpdate, SYNC);
      %%assert(pVar->Status == pvStatOK);

      Status = pvGet(AttenuationCoefficient);
      %%assert(pVar->Status == pvStatOK);

      %{
        data = (struct ImcaXiaFiltersData *)pVar->FiltersData;
        imcaXiaFiltersUpdateAttenuation(data, pVar->AttenuationCoefficient);
        idx = imcaXiaFiltersFindNearestAttenuation(data,
          pVar->RequestedAttenuation);
        assert(idx >= 0);
        pVar->RequestedThickness = data->filterSets[idx].thickness;
        for (i = 0; i < NUM_FILTERS; i++)
          pVar->RequestedFilterStates[i] = data->filterSets[idx].states[i];
      }%

      IgnoreNextRequestedThicknessChange = 1;
      Status = pvPut(RequestedThickness);
      %%assert(pVar->Status == pvStatOK);

      %{
        pVar->AllStatesSame = imcaXiaAreAllStatesSame(
            pVar->RequestedFilterStates, NUM_FILTERS);
      }%

      if (AllStatesSame) {
        RequestedAllFiltersState = RequestedFilterStates[0];
        Status = pvPut(RequestedAllFiltersState, SYNC);
        %%assert(pVar->Status == pvStatOK);
      } else {
        for (Index = 0; Index < NUM_FILTERS; Index++) {
          Status = pvPut(RequestedFilterStates[Index], SYNC);
          %%assert(pVar->Status == pvStatOK);
        }
      }

      RequestedAttenuationBusy = 0;
      Status = pvPut(RequestedAttenuationBusy);
      %%assert(pVar->Status == pvStatOK);
    } state waitForSomethingToDo
  }

  state handleRequestedThicknessChanged {
    when (IgnoreNextRequestedThicknessChange) {
      IgnoreNextRequestedThicknessChange = 0;
      RequestedThicknessBusy = 0;
      Status = pvPut(RequestedThicknessBusy);
      %%assert(pVar->Status == pvStatOK);
    } state waitForSomethingToDo
    when (!IgnoreNextRequestedThicknessChange) {
      %%struct ImcaXiaFiltersData *data;
      %%int i, idx;

      Status = pvGet(RequestedThickness);
      %%assert(pVar->Status == pvStatOK);

      AttenuationCoefficientUpdate = 1;
      Status = pvPut(AttenuationCoefficientUpdate, SYNC);
      %%assert(pVar->Status == pvStatOK);

      Status = pvGet(AttenuationCoefficient);
      %%assert(pVar->Status == pvStatOK);

      %{
        data = (struct ImcaXiaFiltersData *)pVar->FiltersData;
        imcaXiaFiltersUpdateAttenuation(data, pVar->AttenuationCoefficient);
        idx = imcaXiaFiltersFindNearestThickness(data,
          pVar->RequestedThickness);
        assert(idx >= 0);
        pVar->RequestedAttenuation = data->filterSets[idx].attenuation;
        for (i = 0; i < NUM_FILTERS; i++)
          pVar->RequestedFilterStates[i] = data->filterSets[idx].states[i];
      }%

      IgnoreNextRequestedAttenuationChange = 1;
      Status = pvPut(RequestedAttenuation);
      %%assert(pVar->Status == pvStatOK);

      %{
        pVar->AllStatesSame = imcaXiaAreAllStatesSame(
            pVar->RequestedFilterStates, NUM_FILTERS);
      }%

      if (AllStatesSame) {
        RequestedAllFiltersState = RequestedFilterStates[0];
        Status = pvPut(RequestedAllFiltersState, SYNC);
        %%assert(pVar->Status == pvStatOK);
      } else {
        for (Index = 0; Index < NUM_FILTERS; Index++) {
          Status = pvPut(RequestedFilterStates[Index], SYNC);
          %%assert(pVar->Status == pvStatOK);
        }
      }

      RequestedThicknessBusy = 0;
      Status = pvPut(RequestedThicknessBusy);
      %%assert(pVar->Status == pvStatOK);
    } state waitForSomethingToDo
  }

  state handleEnergyMoved {
    when (MaintainRequestedAttenuation) {
    } state triggerRequestedAttenuationChange
    when (!MaintainRequestedAttenuation) {
    } state waitForSomethingToDo
  }

  state handleMaintainRequestedAttenuationEnabled {
    when () {
    } state triggerRequestedAttenuationChange
  }

  state triggerRequestedAttenuationChange {
    when () {
      RequestedAttenuationUpdate = 1;
      Status = pvPut(RequestedAttenuationUpdate);
      %%assert(pVar->Status == pvStatOK);
    } state waitForSomethingToDo
  }
}


exit {
  %%free((void *)pVar->FiltersData);
}
