#include <math.h>

#include <iocsh.h>
#include <epicsExport.h>
#include <registryFunction.h>

#include <imca-ca-utils.h>
#include <imca-epics-unit.h>

#define PV_PREFIX_A "x7ida:"
#define PV_PREFIX_B "x7idb:"

#define PV_REQUESTED_ATTENUATION "FltrRak1:RqsAtten"
#define PV_ATTENUATION "FltrRak1:Atten"
#define PV_MAINTAIN_ATTENUATION "FltrRak1:MntnRqsAtten"
#define PV_REQUESTED_THICKNESS "FltrRak1:RqsThickSum"
#define PV_THICKNESS "FltrRak1:ThickSum"
#define PV_ENERGY "EnergyInEv"
#define PV_ENERGY_MOVING "EnergyMoving"

static EpicsUnitContext euc;

static void testSetAttenuation01() {
  /* Basic. */
  int status;
  dbr_double_t value;

  status = imcaSetEpicsPvToDouble(PV_PREFIX_A PV_ENERGY, 12398.424);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);

  status = imcaSetEpicsPvToDouble(PV_PREFIX_B PV_REQUESTED_ATTENUATION, 50.0);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);

  value = 49.4;
  status = imcaWaitForEpicsPvToHaveValueWithDelta(PV_PREFIX_B PV_ATTENUATION,
    DBR_DOUBLE, &value, 0.1);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);

  value = 180.0;
  status = imcaWaitForEpicsPvToHaveValueWithDelta(PV_PREFIX_B PV_THICKNESS,
    DBR_DOUBLE, &value, 0.001);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);

  status = imcaSetEpicsPvToDouble(PV_PREFIX_B PV_REQUESTED_ATTENUATION, 25.0);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);

  value = 26.1;
  status = imcaWaitForEpicsPvToHaveValueWithDelta(PV_PREFIX_B PV_ATTENUATION,
    DBR_DOUBLE, &value, 0.1);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);

  value = 80.0;
  status = imcaWaitForEpicsPvToHaveValueWithDelta(PV_PREFIX_B PV_THICKNESS,
    DBR_DOUBLE, &value, 0.001);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);

  status = imcaSetEpicsPvToDouble(PV_PREFIX_B PV_REQUESTED_ATTENUATION, 0.0);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);

  value = 0.0;
  status = imcaWaitForEpicsPvToHaveValueWithDelta(PV_PREFIX_B PV_ATTENUATION,
    DBR_DOUBLE, &value, 0.1);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);

  value = 0.0;
  status = imcaWaitForEpicsPvToHaveValueWithDelta(PV_PREFIX_B PV_THICKNESS,
    DBR_DOUBLE, &value, 0.001);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);

  status = imcaSetEpicsPvToDouble(PV_PREFIX_B PV_REQUESTED_ATTENUATION,
    100.0);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);

  value = 100.0;
  status = imcaWaitForEpicsPvToHaveValueWithDelta(PV_PREFIX_B PV_ATTENUATION,
    DBR_DOUBLE, &value, 0.1);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);

  value = 2400.0;
  status = imcaWaitForEpicsPvToHaveValueWithDelta(PV_PREFIX_B PV_THICKNESS,
    DBR_DOUBLE, &value, 0.001);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);
}

static void testMaintainAttenuation01() {
  /* Basic. */
  int status;
  dbr_double_t value;

  status = imcaSetEpicsPvToEnum(PV_PREFIX_B PV_MAINTAIN_ATTENUATION, 1);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);

  status = imcaSetEpicsPvToDouble(PV_PREFIX_A PV_ENERGY, 12398.424);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);

  status = imcaSetEpicsPvToDouble(PV_PREFIX_B PV_REQUESTED_ATTENUATION, 75.0);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);

  epicsThreadSleep(0.5);

  status = imcaSetEpicsPvToDouble(PV_PREFIX_A PV_ENERGY, 11000.0);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);

  status = imcaSetEpicsPvToEnum(PV_PREFIX_A PV_ENERGY_MOVING, 1);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);
  epicsThreadSleep(0.25);
  status = imcaSetEpicsPvToEnum(PV_PREFIX_A PV_ENERGY_MOVING, 0);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);

  value = 75.1;
  status = imcaWaitForEpicsPvToHaveValueWithDelta(PV_PREFIX_B PV_ATTENUATION,
    DBR_DOUBLE, &value, 0.1);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);

  value = 260.0;
  status = imcaWaitForEpicsPvToHaveValueWithDelta(PV_PREFIX_B PV_THICKNESS,
    DBR_DOUBLE, &value, 0.001);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);
}
static void testMaintainAttenuation02() {
  /* Enabling maintain attenuation should trigger set for current energy. */
  int status;
  dbr_double_t value;

  status = imcaSetEpicsPvToEnum(PV_PREFIX_B PV_MAINTAIN_ATTENUATION, 0);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);

  epicsThreadSleep(0.25);

  status = imcaSetEpicsPvToDouble(PV_PREFIX_A PV_ENERGY, 12398.424);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);

  status = imcaSetEpicsPvToDouble(PV_PREFIX_B PV_REQUESTED_ATTENUATION, 75.0);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);

  epicsThreadSleep(0.5);

  status = imcaSetEpicsPvToDouble(PV_PREFIX_A PV_ENERGY, 11000.0);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);

  status = imcaSetEpicsPvToEnum(PV_PREFIX_A PV_ENERGY_MOVING, 1);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);
  epicsThreadSleep(0.25);
  status = imcaSetEpicsPvToEnum(PV_PREFIX_A PV_ENERGY_MOVING, 0);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);

  value = 86.2;
  status = imcaWaitForEpicsPvToHaveValueWithDelta(PV_PREFIX_B PV_ATTENUATION,
    DBR_DOUBLE, &value, 0.1);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);

  value = 370.0;
  status = imcaWaitForEpicsPvToHaveValueWithDelta(PV_PREFIX_B PV_THICKNESS,
    DBR_DOUBLE, &value, 0.001);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);

  status = imcaSetEpicsPvToEnum(PV_PREFIX_B PV_MAINTAIN_ATTENUATION, 1);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);

  value = 75.1;
  status = imcaWaitForEpicsPvToHaveValueWithDelta(PV_PREFIX_B PV_ATTENUATION,
    DBR_DOUBLE, &value, 0.1);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);

  value = 260.0;
  status = imcaWaitForEpicsPvToHaveValueWithDelta(PV_PREFIX_B PV_THICKNESS,
    DBR_DOUBLE, &value, 0.001);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);
}
static void testMaintainAttenuation03() {
  /* Set attenuation immediately after an energy move finishes. */
  int i;
  int status;
  dbr_double_t value;

  /* Since this test is sensitive to timing, it might not always fail.  So,
     repeat it 10 times to be sure that it's working correctly. */
  for (i = 0; i < 10; i++) {
    status = imcaSetEpicsPvToEnum(PV_PREFIX_B PV_MAINTAIN_ATTENUATION, 1);
    failIfIntNotEqual(&euc, ECA_NORMAL, status);
    status = imcaSetEpicsPvToDouble(PV_PREFIX_A PV_ENERGY, 12398.424);
    failIfIntNotEqual(&euc, ECA_NORMAL, status);
    status = imcaSetEpicsPvToDouble(PV_PREFIX_B PV_REQUESTED_ATTENUATION,
      94.0);
    failIfIntNotEqual(&euc, ECA_NORMAL, status);

    epicsThreadSleep(0.25);

    status = imcaSetEpicsPvToEnum(PV_PREFIX_A PV_ENERGY_MOVING, 1);
    failIfIntNotEqual(&euc, ECA_NORMAL, status);
    status = imcaSetEpicsPvToDouble(PV_PREFIX_A PV_ENERGY, 12291.913);
    failIfIntNotEqual(&euc, ECA_NORMAL, status);
    status = imcaSetEpicsPvToEnum(PV_PREFIX_A PV_ENERGY_MOVING, 0);
    failIfIntNotEqual(&euc, ECA_NORMAL, status);
    status = imcaSetEpicsPvToDouble(PV_PREFIX_B PV_REQUESTED_ATTENUATION,
      55.0);
    failIfIntNotEqual(&euc, ECA_NORMAL, status);

    value = 55.7028;
    status = imcaWaitForEpicsPvToHaveValueWithDeltaAndTimeout(
      PV_PREFIX_B PV_ATTENUATION, DBR_DOUBLE, &value, 0.1, 1.0);
    failIfIntNotEqual(&euc, ECA_NORMAL, status);
    if (status != ECA_NORMAL) break;
  }
}

static void testSetThickness01() {
  /* Basic. */
  int status;
  dbr_double_t value;

  status = imcaSetEpicsPvToDouble(PV_PREFIX_A PV_ENERGY, 12398.424);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);

  status = imcaSetEpicsPvToDouble(PV_PREFIX_B PV_REQUESTED_THICKNESS, 180.0);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);

  value = 180.0;
  status = imcaWaitForEpicsPvToHaveValueWithDelta(PV_PREFIX_B PV_THICKNESS,
    DBR_DOUBLE, &value, 0.001);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);

  status = imcaSetEpicsPvToDouble(PV_PREFIX_B PV_REQUESTED_THICKNESS, 200.0);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);

  value = 200.0;
  status = imcaWaitForEpicsPvToHaveValueWithDelta(PV_PREFIX_B PV_THICKNESS,
    DBR_DOUBLE, &value, 0.001);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);

  value = 53.1;
  status = imcaWaitForEpicsPvToHaveValueWithDelta(
    PV_PREFIX_B PV_REQUESTED_ATTENUATION, DBR_DOUBLE, &value, 0.1);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);

  status = imcaSetEpicsPvToDouble(PV_PREFIX_B PV_REQUESTED_THICKNESS, 201.0);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);

  value = 200.0;
  status = imcaWaitForEpicsPvToHaveValueWithDelta(PV_PREFIX_B PV_THICKNESS,
    DBR_DOUBLE, &value, 0.001);
  failIfIntNotEqual(&euc, ECA_NORMAL, status);
}

static void runAllTests() {
  runTest(&euc, testSetAttenuation01);

  runTest(&euc, testMaintainAttenuation01);
  runTest(&euc, testMaintainAttenuation02);
  runTest(&euc, testMaintainAttenuation03);

  runTest(&euc, testSetThickness01);
}

void imcaXiaPfcu4x2TestRunner() {
  runTests(&euc, runAllTests);
}

static const iocshArg *const imcaXiaPfcu4x2TestRunnerArgs[] = {};
static const iocshFuncDef imcaXiaPfcu4x2TestRunnerDef = {
  "imcaXiaPfcu4x2TestRunner", 0, imcaXiaPfcu4x2TestRunnerArgs
};
static void imcaXiaPfcu4x2TestRunnerCall(const iocshArgBuf *args) {
  imcaXiaPfcu4x2TestRunner();
}
static void imcaXiaPfcu4x2TestRegistrar(void) {
  static int isFirstTime = 1;
  if (!isFirstTime) return;
  isFirstTime = 0;
  iocshRegister(&imcaXiaPfcu4x2TestRunnerDef, imcaXiaPfcu4x2TestRunnerCall);
}
epicsExportRegistrar(imcaXiaPfcu4x2TestRegistrar);
