#!/usr/bin/python2

#
# I simulate an XIA PFCU-4 filter controller.
#
# For usage, see my printUsage function or run me with my -h option.
#

import getopt
import asynchat
import asyncore
import os
import re
import signal
import socket
import sys
import time

DEFAULT_PORT = 50439

def main(args):
  (port, isDebugOutputEnabled) = parseCommandLineArgs(args)
  d = ConnectionDispatcher(port, isDebugOutputEnabled)
  asyncore.loop()

def getProgramName(args=None):
  if args == None: args = sys.argv
  if len(args) == 0 or args[0] == "-c": return "PROGRAM_NAME"
  return os.path.basename(args[0])

def initializeSignalHandlers():
  signalToNameMap = {
    signal.SIGHUP:"SIGHUP",
    signal.SIGINT:"SIGINT",
    signal.SIGPIPE:"SIGPIPE",
    signal.SIGALRM:"SIGALRM",
    signal.SIGTERM:"SIGTERM",
    signal.SIGUSR1:"SIGUSR1",
    signal.SIGUSR2:"SIGUSR2"
  }

  def signalHandler(signum, frame):
    if signum in [signal.SIGINT, signal.SIGTERM]:
      print
      print "%s caught. Exiting." % signalToNameMap[signum]
      sys.exit(0)
    else:
      print >> sys.stderr
      print >> sys.stderr, "%s caught. Exiting." % signalToNameMap[signum]
      sys.exit(1)

  for each in signalToNameMap.keys():
    signal.signal(each, signalHandler)

def parseCommandLineArgs(args):
  (options, extra) = getopt.getopt(args[1:], "dp:h", ["debug", "port=",
    "help"])

  port = DEFAULT_PORT
  isDebugOutputEnabled = False

  for eachOptName, eachOptValue in options:
    if eachOptName in ("-d", "--debug"):
      isDebugOutputEnabled = True
    elif eachOptName in ("-p", "--port"):
      port = int(eachOptValue)
    elif eachOptName in ("-h", "--help"):
      printUsage(sys.stdout)
      sys.exit(0)

  if len(extra) > 0:
    print >> sys.stderr, "Error: unexpected command line argument \"%s\"" % \
      extra[0]
    printUsage(sys.stderr)
    sys.exit(1)

  return (port, isDebugOutputEnabled)

def printUsage(outStream):
  print >> outStream, """\
Usage: %s [-dph]

Options:
  -d,--debug        Print debug messages to stderr
  -p,--port=NUMBER  Listen on the specified port NUMBER for incoming
                    connections (default: %d)
  -h,--help         Print usage message and exit\
""" % (getProgramName(), DEFAULT_PORT)


class ConnectionDispatcher(asyncore.dispatcher):
  def __init__(self, port, debugOutputEnabled=False):
    asyncore.dispatcher.__init__(self)
    self.port = port
    self.debugOutputEnabled = debugOutputEnabled
    self.devices = []
    for each in xrange(16): self.devices.append(XiaPfcu4())
    self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
    self.set_reuse_addr()
    self.bind(("", port))
    self.listen(5)

  def handle_accept(self):
    ConnectionSession(self.accept(), self.devices, self.debugOutputEnabled)


class ConnectionSession(asynchat.async_chat):
  FaultStatusRe = re.compile(r"!PFCU(?P<id>[0-9]{2}) F$", re.I)
  WriteFilterRe = re.compile(r"!PFCU(?P<id>[0-9]{2}) W *" + \
    r"(?P<a>[01])(?P<b>[01])(?P<c>[01])(?P<d>[01])$", re.I)
  InsertFilterRe = re.compile(r"!PFCU(?P<id>[0-9]{2}) I *" + \
    r"(?P<a>[1-4])(?P<b>[1-4])*(?P<c>[1-4])*(?P<d>[1-4])*$", re.I)
  RemoveFilterRe = re.compile(r"!PFCU(?P<id>[0-9]{2}) R *" + \
    r"(?P<a>[1-4])(?P<b>[1-4])*(?P<c>[1-4])*(?P<d>[1-4])*$", re.I)
  IdRe = re.compile(r"!PFCU(?P<id>[0-9]{2})", re.I)

  def __init__(self, (conn, addr), devices, debugOutputEnabled=False):
    asynchat.async_chat.__init__(self, conn)
    self.debugOutputEnabled = debugOutputEnabled
    self.set_terminator("\r")
    self.outputTerminator = ";\r\n"
    self.devices = devices
    self.buffer = ""

  def collect_incoming_data(self, data):
    self.buffer = self.buffer + data

  def found_terminator(self):
    data = self.buffer
    self.buffer = ""
    if self.debugOutputEnabled:
      print >> sys.stderr, "< \"%s\"" % \
        (data + self.get_terminator()).encode("string_escape")
    self.handleClientRequest(data)

  def handleClientRequest(self, request):
    request = request.strip()
    match = self.FaultStatusRe.match(request)
    if match != None:
      id = int(match.group("id"))
      device = self.devices[id]
      self.sendClientResponse("%%PFCU%02d OK %d%d%d%d" % \
        (id, device.getFilterState(0), device.getFilterState(1),
         device.getFilterState(2), device.getFilterState(3)))
      return
    match = self.WriteFilterRe.match(request)
    if match != None:
      id = int(match.group("id"))
      device = self.devices[id]
      for each in xrange(4):
        device.setFilterState(each, int(match.group("abcd"[each])))
      self.sendClientResponse("%%PFCU%02d OK %d%d%d%d DONE" % \
        (id, device.getFilterState(0), device.getFilterState(1),
         device.getFilterState(2), device.getFilterState(3)))
      return
    match = self.InsertFilterRe.match(request)
    if match != None:
      id = int(match.group("id"))
      device = self.devices[id]
      for each in xrange(4):
        filterId = match.group("abcd"[each])
        if filterId != None: device.setFilterState(int(filterId) - 1, 1)
      self.sendClientResponse("%%PFCU%02d OK %d%d%d%d DONE" % \
        (id, device.getFilterState(0), device.getFilterState(1),
         device.getFilterState(2), device.getFilterState(3)))
      return
    match = self.RemoveFilterRe.match(request)
    if match != None:
      id = int(match.group("id"))
      device = self.devices[id]
      for each in xrange(4):
        filterId = match.group("abcd"[each])
        if filterId != None: device.setFilterState(int(filterId) - 1, 0)
      self.sendClientResponse("%%PFCU%02d OK %d%d%d%d DONE" % \
        (id, device.getFilterState(0), device.getFilterState(1),
         device.getFilterState(2), device.getFilterState(3)))
      return

    print >> sys.stderr, "error: unknown command \"%s\"" % \
      request.encode("string_escape")
    match = self.IdRe.match(request)
    if match != None:
      id = int(match.group("id"))
      device = self.devices[id]
      self.sendClientResponse("%%PFCU%02d ERROR: Invalid command" % id)

  def sendClientResponse(self, response=""):
    data = response + self.outputTerminator
    self.push(data)
    if self.debugOutputEnabled:
      print >> sys.stderr, "> \"%s\"" % data.encode("string_escape")


class XiaPfcu4:

  def __init__(self):
    self.filterStates = [0, 0, 0, 0]

  def getFilterState(self, filter):
    return self.filterStates[filter]

  def setFilterState(self, filter, state):
    self.filterStates[filter] = state


if __name__ == '__main__':
  try:
    initializeSignalHandlers()
    main(sys.argv)
  except Exception, e:
    if isinstance(e, SystemExit):
      raise e
    else:
      print >> sys.stderr
      print >> sys.stderr, "Error: %s" % e
      sys.exit(1)
